import { credentialsProvider } from "@tidal-music/auth";
import { FlacStreamTagger } from "flac-stream-tagger";
import Ffmpeg from "fluent-ffmpeg";
import ffmpegPath from "ffmpeg-static";
import { parse } from "dasha";

import { makeDecipher } from "./decrypt";

Ffmpeg.setFfmpegPath(ffmpegPath as string);

export default class {
    private _authHeaders: { [key: string]: string }
    private _userId: string | undefined;
    private _userCountry: string;
    private _userLocale: string;

    constructor(authHeaders: { [key: string]: string }) {
        this._authHeaders = authHeaders;
        this._userCountry = "WW"; // Default to Worldwide
        this._userLocale = (process.env.LC_ALL ||
            process.env.LANG ||
            process.env.LANGUAGE ||
            Intl.DateTimeFormat().resolvedOptions().locale ||
            "en_US").replaceAll("-", "_").split(".")[0];
    }

    async fetch(url: string, opts: { [key: string]: any } = {}) {
        const parsedUrl = new URL(`https://desktop.tidal.com/v1${url}`);

        parsedUrl.searchParams.set("countryCode", this._userCountry);
        parsedUrl.searchParams.set("locale", this._userLocale);
        parsedUrl.searchParams.set("deviceType", "DESKTOP");

        const response = await fetch(parsedUrl, {
            headers: {
                ...this._authHeaders,
                "x-tidal-token": "mhPVJJEBNRzVjr2p",
            },
            referrer: "https://desktop.tidal.com/",
            ...opts
        })

        if (!response.ok) {
            throw new Error(`Failed to fetch ${url}: ${response.status} ${response.statusText}`)
        }

        return await response.json() as Object;
    }

    async init() {
        const creds = await credentialsProvider.getCredentials()

        this._userId = creds.userId

        const { countryCode } = await this.fetch("/country") as { countryCode: string }
        this._userCountry = countryCode;
    }

    async fetchTrack(id: number) {
        const audio = await this.fetch(`/tracks/${id}/playbackinfo?audioquality=HI_RES_LOSSLESS&playbackmode=STREAM&assetpresentation=FULL`)

        return audio as {
            trackId: number,
            audioPresentation: string,
            audioMode: string,
            audioQuality: string,
            manifestMimeType: string,
            manifestHash: string,
            manifest: string,
            albumReplayGain: number,
            albumPeakAmplitude: number,
            trackReplayGain: number,
            trackPeakAmplitude: number,
            bitDepth: number,
            sampleRate: number,
            status?: number,
        }
    }

    async fetchTracks() {
        const tracks = await this.fetch(`/users/${this._userId}/favorites/tracks?offset=0&limit=10000&order=DATE&orderDirection=DESC`);

        return tracks;
    }

    async downloadFlac(manifestMimeType: string, manifest: string) {
        if (manifestMimeType === "application/dash+xml") {
            const parsedManifest = await parse(Buffer.from(manifest, "base64").toString("utf-8"), "https://sp-ad-cf.audio.tidal.com")

            const segmentBuffers = [];

            for await (const track of parsedManifest.tracks.all) {
                for await (const segment of track.segments) {
                    const req = await fetch(segment.url)
                    const content = await req.arrayBuffer()
                    const contentBuffer = Buffer.from(content)
                    segmentBuffers.push(contentBuffer)
                }
            }

            return { buffer: Buffer.concat(segmentBuffers), mimeType: "audio/m4a" }
        } else if (manifestMimeType === "application/vnd.tidal.bts") {
            const data = JSON.parse(Buffer.from(manifest, "base64").toString("utf-8")) as {
                "mimeType": string,
                "codecs": string,
                "encryptionType": string,
                "keyId": string,
                "urls": string[]
            }

            const decryptor = makeDecipher({ keyId: data.keyId, encryptionType: data.encryptionType });

            const flac = await fetch(data.urls[0])

            const flacBuffer = await flac.arrayBuffer()

            const decryptedBuffer = decryptor ? Buffer.concat([decryptor.update(Buffer.from(flacBuffer)), decryptor.final()]) : Buffer.from(flacBuffer);

            return { buffer: decryptedBuffer, mimeType: "audio/flac" };
        }

        return { buffer: new ArrayBuffer(0), mimeType: "audio/unknown" } // TODO: Handle other mime types
    }

    async tagFlac(id: number, audioBuffer: Buffer<ArrayBuffer> | ArrayBuffer) {
        const trackReq = await fetch(`https://desktop.tidal.com/v1/tracks/${id}/?countryCode=US`, {
            headers: this._authHeaders
        })

        const track = await trackReq.json() as {
            title: string,
            replayGain: number,
            peak: number,
            streamStartDate: string,
            trackNumber: number,
            volumeNumber: number, // discNumber
            copyright: string,
            bpm: number,
            url: string, // comment
            isrc: string,
            artists: [
                {
                    name: string,
                }
            ],
            album: {
                id: number,
                title: string,
                cover: string,
            },
            status?: string,
        };


        const albumReq = await fetch(`https://desktop.tidal.com/v1/albums/${track.album.id}/?countryCode=US`, {
            headers: this._authHeaders
        })

        const album = await albumReq.json() as {
            title: string,
            numberOfTracks: number,
            releaseDate: string,
            cover: string,
            upc: string,
            artists: [
                {
                    name: string,
                }
            ],
        }

        const albumArtReq = await fetch(`https://resources.tidal.com/images/${album?.cover.replaceAll("-", "/") || ""}/1280x1280.jpg`)
        const albumArt = albumArtReq.ok ? await albumArtReq.arrayBuffer() : new ArrayBuffer(0);


        return FlacStreamTagger.fromBuffer(Buffer.from(Buffer.isBuffer(audioBuffer) ? audioBuffer : Buffer.from(audioBuffer)), {
            tagMap: {
                title: track?.title || "",
                trackNumber: (track?.trackNumber || 0).toString(),
                discNumber: (track?.volumeNumber || 0).toString(),
                bpm: (track?.bpm || 0).toString(),
                date: track?.streamStartDate || "",
                copyright: track?.copyright || "",
                REPLAYGAIN_TRACK_GAIN: (track?.replayGain || 0).toString(),
                REPLAYGAIN_TRACK_PEAK: (track?.peak || 0).toString(),
                comment: track?.url || "",
                isrc: track?.isrc || "",
                upc: album?.upc || "",
                artist: track?.artists.map((a) => a.name) || [],
                album: album?.title || "",
                albumArtist: album?.artists.map((a) => a.name) || [],
                totalTracks: (album?.numberOfTracks || 0).toString(),
                year: album?.releaseDate.split("-")[0] || "",
            },
            picture: {
                buffer: Buffer.from(albumArt),
            }
        }).toBuffer()
    }

}